/**
 * @file   EnergyComponentPrinter.cpp
 * @author Thomas Dresselhaus <t.dresselhaus at wwu.de>
 *
 * @date   5. Januar 2015, 14:19
 * @copyright \n
 *  This file is part of the program Serenity.\n\n
 *  Serenity is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as
 *  published by the Free Software Foundation, either version 3 of
 *  the License, or (at your option) any later version.\n\n
 *  Serenity is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.\n\n
 *  You should have received a copy of the GNU Lesser General
 *  Public License along with Serenity.
 *  If not, see <http://www.gnu.org/licenses/>.\n
 */
/* Include Class Header*/
#include "energies/EnergyComponentPrinter.h"
/* Include Serenity Internal Headers */
#include "energies/EnergyContributions.h"
#include "misc/SerenityError.h"
/* Include Std and External Headers */
#include <stdio.h>

namespace Serenity {

const std::vector<ENERGY_CONTRIBUTIONS> EnergyComponentPrinter::ENERGY_CONTRIBUTIONS_VECTOR = {
    ENERGY_CONTRIBUTIONS::ELECTRON_KINETIC,
    ENERGY_CONTRIBUTIONS::ELECTRON_ECP,
    ENERGY_CONTRIBUTIONS::ELECTRON_NUCLEUS_COULOMB,
    ENERGY_CONTRIBUTIONS::ELECTRON_ELECTRON_COULOMB,
    ENERGY_CONTRIBUTIONS::HF_ELECTRON_ELECTRON_EXCHANGE,
    ENERGY_CONTRIBUTIONS::KS_DFT_ELECTRON_ELECTRON_EXCHANGE,
    ENERGY_CONTRIBUTIONS::ELECTRON_ELECTRON_CORRELATION,
    ENERGY_CONTRIBUTIONS::NUCLEUS_NUCLEUS_COULOMB,
    ENERGY_CONTRIBUTIONS::KS_DFT_DISPERSION_CORRECTION,
    ENERGY_CONTRIBUTIONS::ONE_ELECTRON_ENERGY,
    ENERGY_CONTRIBUTIONS::HF_TWO_ELECTRON_ENERGY,
    ENERGY_CONTRIBUTIONS::KS_DFT_EXCHANGE_CORRELATION_ENERGY,
    ENERGY_CONTRIBUTIONS::KS_DFT_EXCHANGE_CORRELATION_NO_HF,
    ENERGY_CONTRIBUTIONS::KS_DFT_EXACT_EXCHANGE,
    ENERGY_CONTRIBUTIONS::KS_DFT_PERTURBATIVE_CORRELATION,
    ENERGY_CONTRIBUTIONS::KS_DFT_TWO_ELECTRON_ENERGY,
    ENERGY_CONTRIBUTIONS::HF_ENERGY,
    ENERGY_CONTRIBUTIONS::KS_DFT_ENERGY,
    ENERGY_CONTRIBUTIONS::FDE_NAD_EXCHANGE,
    ENERGY_CONTRIBUTIONS::FDE_NAD_CORRELATION,
    ENERGY_CONTRIBUTIONS::FDE_NAD_XC,
    ENERGY_CONTRIBUTIONS::FDE_NAD_EXACT_EXCHANGE,
    ENERGY_CONTRIBUTIONS::FDE_NAD_KINETIC,
    ENERGY_CONTRIBUTIONS::FDE_NAD_DISP,
    ENERGY_CONTRIBUTIONS::FDE_NUCLEI_ENV_ELECTRONS_COULOMB,
    ENERGY_CONTRIBUTIONS::FDE_NUCLEI_ENV_NUCLEI_COULOMB,
    ENERGY_CONTRIBUTIONS::FDE_ELECTRONS_ENV_ELECTRONS_COULOMB,
    ENERGY_CONTRIBUTIONS::FDE_ELECTRONS_ENV_NUCLEI_COULOMB,
    ENERGY_CONTRIBUTIONS::FDE_ELECTROSTATIC_INTERACTIONS,
    ENERGY_CONTRIBUTIONS::FDE_INTERACTION_ENERGY,
    ENERGY_CONTRIBUTIONS::FDE_EMBEDDED_KS_DFT_ENERGY,
    ENERGY_CONTRIBUTIONS::FDE_EMBEDDED_HF_ENERGY,
    ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_NAD_EXCHANGE,
    ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_NAD_CORRELATION,
    ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_NAD_XC,
    ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_NAD_EXACT_EXCHANGE,
    ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_NAD_KINETIC,
    ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_NAD_DISP,
    ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_NUCLEI_ENV_ELECTRONS_COULOMB,
    ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_ELECTRONS_ENV_ELECTRONS_COULOMB,
    ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_ELECTRONS_ENV_NUCLEI_COULOMB,
    ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_ELECTROSTATIC_INTERACTIONS,
    ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_INTERACTION_ENERGY,
    ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_EMBEDDED_KS_DFT_ENERGY,
    ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_EMBEDDED_HF_ENERGY,
    ENERGY_CONTRIBUTIONS::FDE_SUPERSYSTEM_ENERGY_DFT_DFT,
    ENERGY_CONTRIBUTIONS::FDE_SUPERSYSTEM_ENERGY_WF_DFT,
    ENERGY_CONTRIBUTIONS::FDE_FROZEN_SUBSYSTEM_ENERGIES,
    ENERGY_CONTRIBUTIONS::FDE_ECP_ENV_ENERGY,
    ENERGY_CONTRIBUTIONS::PBE_LINEAR_CORRECTION,
    ENERGY_CONTRIBUTIONS::PBE_SUPERSYSTEM_ENERGY_WFT_DFT,
    ENERGY_CONTRIBUTIONS::PBE_SUPERSYSTEM_ENERGY_DFT_DFT,
    ENERGY_CONTRIBUTIONS::MP2_CORRECTION,
    ENERGY_CONTRIBUTIONS::CCSD_CORRECTION,
    ENERGY_CONTRIBUTIONS::TRIPLES_CORRECTION,
    ENERGY_CONTRIBUTIONS::MP2_TOTAL_ENERGY,
    ENERGY_CONTRIBUTIONS::CCSD_TOTAL_ENERGY,
    ENERGY_CONTRIBUTIONS::CCSD_T_TOTAL_ENERGY,
    ENERGY_CONTRIBUTIONS::SOLVATION_ENERGY,
};

const std::map<ENERGY_CONTRIBUTIONS, EnergyContributionChildren> EnergyComponentPrinter::ENERGY_CONTRIBUTIONS_CHILDREN_MAP = {
    {ENERGY_CONTRIBUTIONS::ELECTRON_KINETIC, {}},
    {ENERGY_CONTRIBUTIONS::ELECTRON_ECP, {}},
    {ENERGY_CONTRIBUTIONS::SOLVATION_ENERGY, {}},
    {ENERGY_CONTRIBUTIONS::ELECTRON_NUCLEUS_COULOMB, {}},
    {ENERGY_CONTRIBUTIONS::ELECTRON_ELECTRON_COULOMB, {}},
    {ENERGY_CONTRIBUTIONS::HF_ELECTRON_ELECTRON_EXCHANGE, {}},
    {ENERGY_CONTRIBUTIONS::KS_DFT_ELECTRON_ELECTRON_EXCHANGE, {}},
    {ENERGY_CONTRIBUTIONS::ELECTRON_ELECTRON_CORRELATION, {}},
    {ENERGY_CONTRIBUTIONS::NUCLEUS_NUCLEUS_COULOMB, {}},
    {ENERGY_CONTRIBUTIONS::KS_DFT_DISPERSION_CORRECTION, {}},
    {ENERGY_CONTRIBUTIONS::KS_DFT_EXACT_EXCHANGE, {}},
    {ENERGY_CONTRIBUTIONS::KS_DFT_PERTURBATIVE_CORRELATION, {}},
    {ENERGY_CONTRIBUTIONS::ONE_ELECTRON_ENERGY,
     {ENERGY_CONTRIBUTIONS::ELECTRON_NUCLEUS_COULOMB, ENERGY_CONTRIBUTIONS::ELECTRON_KINETIC}},
    {ENERGY_CONTRIBUTIONS::HF_TWO_ELECTRON_ENERGY,
     {ENERGY_CONTRIBUTIONS::ELECTRON_ELECTRON_COULOMB, ENERGY_CONTRIBUTIONS::HF_ELECTRON_ELECTRON_EXCHANGE}},
    {ENERGY_CONTRIBUTIONS::KS_DFT_EXCHANGE_CORRELATION_NO_HF,
     {ENERGY_CONTRIBUTIONS::KS_DFT_ELECTRON_ELECTRON_EXCHANGE, ENERGY_CONTRIBUTIONS::ELECTRON_ELECTRON_CORRELATION}},
    {ENERGY_CONTRIBUTIONS::KS_DFT_EXCHANGE_CORRELATION_ENERGY,
     {ENERGY_CONTRIBUTIONS::KS_DFT_EXCHANGE_CORRELATION_NO_HF, ENERGY_CONTRIBUTIONS::KS_DFT_EXACT_EXCHANGE,
      ENERGY_CONTRIBUTIONS::KS_DFT_PERTURBATIVE_CORRELATION}},
    {ENERGY_CONTRIBUTIONS::KS_DFT_TWO_ELECTRON_ENERGY,
     {ENERGY_CONTRIBUTIONS::KS_DFT_EXCHANGE_CORRELATION_ENERGY, ENERGY_CONTRIBUTIONS::ELECTRON_ELECTRON_COULOMB}},
    {ENERGY_CONTRIBUTIONS::HF_ENERGY,
     {ENERGY_CONTRIBUTIONS::ONE_ELECTRON_ENERGY, ENERGY_CONTRIBUTIONS::HF_TWO_ELECTRON_ENERGY,
      ENERGY_CONTRIBUTIONS::NUCLEUS_NUCLEUS_COULOMB, ENERGY_CONTRIBUTIONS::SOLVATION_ENERGY}},
    {ENERGY_CONTRIBUTIONS::KS_DFT_ENERGY,
     {ENERGY_CONTRIBUTIONS::ONE_ELECTRON_ENERGY, ENERGY_CONTRIBUTIONS::KS_DFT_TWO_ELECTRON_ENERGY,
      ENERGY_CONTRIBUTIONS::NUCLEUS_NUCLEUS_COULOMB, ENERGY_CONTRIBUTIONS::KS_DFT_DISPERSION_CORRECTION,
      ENERGY_CONTRIBUTIONS::SOLVATION_ENERGY}},
    {ENERGY_CONTRIBUTIONS::FDE_NAD_EXCHANGE, {}},
    {ENERGY_CONTRIBUTIONS::FDE_NAD_CORRELATION, {}},
    {ENERGY_CONTRIBUTIONS::FDE_MP2_INT_ENERGY, {}},
    {ENERGY_CONTRIBUTIONS::FDE_NAD_XC, {ENERGY_CONTRIBUTIONS::FDE_NAD_EXCHANGE, ENERGY_CONTRIBUTIONS::FDE_NAD_CORRELATION}},
    {ENERGY_CONTRIBUTIONS::FDE_NAD_EXACT_EXCHANGE, {}},
    {ENERGY_CONTRIBUTIONS::FDE_NAD_KINETIC, {}},
    {ENERGY_CONTRIBUTIONS::FDE_NAD_DISP, {}},
    {ENERGY_CONTRIBUTIONS::FDE_NUCLEI_ENV_ELECTRONS_COULOMB, {}},
    {ENERGY_CONTRIBUTIONS::FDE_NUCLEI_ENV_NUCLEI_COULOMB, {}},
    {ENERGY_CONTRIBUTIONS::FDE_ELECTRONS_ENV_NUCLEI_COULOMB, {}},
    {ENERGY_CONTRIBUTIONS::FDE_ELECTRONS_ENV_ELECTRONS_COULOMB, {}},

    {ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_NAD_XC, {}},
    {ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_NAD_KINETIC, {}},
    {ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_NAD_DISP, {}},
    {ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_NUCLEI_ENV_ELECTRONS_COULOMB, {}},
    {ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_NUCLEI_ENV_NUCLEI_COULOMB, {}},
    {ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_ELECTRONS_ENV_NUCLEI_COULOMB, {}},
    {ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_ELECTRONS_ENV_ELECTRONS_COULOMB, {}},

    {ENERGY_CONTRIBUTIONS::FDE_ECP_ENV_ENERGY, {}},
    {ENERGY_CONTRIBUTIONS::FDE_ELECTROSTATIC_INTERACTIONS,
     {ENERGY_CONTRIBUTIONS::FDE_ELECTRONS_ENV_ELECTRONS_COULOMB, ENERGY_CONTRIBUTIONS::FDE_ELECTRONS_ENV_NUCLEI_COULOMB,
      ENERGY_CONTRIBUTIONS::FDE_NUCLEI_ENV_ELECTRONS_COULOMB, ENERGY_CONTRIBUTIONS::FDE_NUCLEI_ENV_NUCLEI_COULOMB}},
    {ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_ELECTROSTATIC_INTERACTIONS,
     {ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_ELECTRONS_ENV_ELECTRONS_COULOMB,
      ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_ELECTRONS_ENV_NUCLEI_COULOMB,
      ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_NUCLEI_ENV_ELECTRONS_COULOMB,
      ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_NUCLEI_ENV_NUCLEI_COULOMB}},
    {ENERGY_CONTRIBUTIONS::FDE_INTERACTION_ENERGY,
     {ENERGY_CONTRIBUTIONS::FDE_ELECTROSTATIC_INTERACTIONS, ENERGY_CONTRIBUTIONS::FDE_NAD_XC,
      ENERGY_CONTRIBUTIONS::FDE_NAD_EXACT_EXCHANGE, ENERGY_CONTRIBUTIONS::FDE_NAD_KINETIC, ENERGY_CONTRIBUTIONS::FDE_NAD_DISP,
      ENERGY_CONTRIBUTIONS::FDE_ECP_ENV_ENERGY, ENERGY_CONTRIBUTIONS::FDE_MP2_INT_ENERGY}},
    {ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_INTERACTION_ENERGY,
     {ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_ELECTROSTATIC_INTERACTIONS, ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_NAD_XC,
      ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_NAD_KINETIC, ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_NAD_DISP,
      ENERGY_CONTRIBUTIONS::FDE_ECP_ENV_ENERGY}},
    {ENERGY_CONTRIBUTIONS::FDE_EMBEDDED_KS_DFT_ENERGY,
     {ENERGY_CONTRIBUTIONS::KS_DFT_ENERGY, ENERGY_CONTRIBUTIONS::FDE_INTERACTION_ENERGY}},
    {ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_EMBEDDED_KS_DFT_ENERGY,
     {ENERGY_CONTRIBUTIONS::KS_DFT_ENERGY, ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_INTERACTION_ENERGY}},
    {ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_EMBEDDED_HF_ENERGY,
     {ENERGY_CONTRIBUTIONS::HF_ENERGY, ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_INTERACTION_ENERGY}},
    {ENERGY_CONTRIBUTIONS::FDE_EMBEDDED_HF_ENERGY,
     {ENERGY_CONTRIBUTIONS::HF_ENERGY, ENERGY_CONTRIBUTIONS::FDE_INTERACTION_ENERGY}},
    {ENERGY_CONTRIBUTIONS::FDE_FROZEN_SUBSYSTEM_ENERGIES, {}},
    {ENERGY_CONTRIBUTIONS::FDE_SUPERSYSTEM_ENERGY_DFT_DFT,
     {ENERGY_CONTRIBUTIONS::KS_DFT_ENERGY, ENERGY_CONTRIBUTIONS::FDE_INTERACTION_ENERGY,
      ENERGY_CONTRIBUTIONS::FDE_FROZEN_SUBSYSTEM_ENERGIES}},
    {ENERGY_CONTRIBUTIONS::FDE_SUPERSYSTEM_ENERGY_WF_DFT,
     {ENERGY_CONTRIBUTIONS::HF_ENERGY, ENERGY_CONTRIBUTIONS::FDE_INTERACTION_ENERGY,
      ENERGY_CONTRIBUTIONS::FDE_FROZEN_SUBSYSTEM_ENERGIES}},
    {ENERGY_CONTRIBUTIONS::PBE_LINEAR_CORRECTION, {}},
    {ENERGY_CONTRIBUTIONS::PBE_SUPERSYSTEM_ENERGY_WFT_DFT,
     {ENERGY_CONTRIBUTIONS::HF_ENERGY, ENERGY_CONTRIBUTIONS::FDE_ELECTROSTATIC_INTERACTIONS, ENERGY_CONTRIBUTIONS::FDE_NAD_XC,
      ENERGY_CONTRIBUTIONS::FDE_NAD_EXACT_EXCHANGE, ENERGY_CONTRIBUTIONS::PBE_LINEAR_CORRECTION,
      ENERGY_CONTRIBUTIONS::FDE_FROZEN_SUBSYSTEM_ENERGIES, ENERGY_CONTRIBUTIONS::FDE_ECP_ENV_ENERGY}},
    {ENERGY_CONTRIBUTIONS::PBE_SUPERSYSTEM_ENERGY_DFT_DFT,
     {ENERGY_CONTRIBUTIONS::KS_DFT_ENERGY, ENERGY_CONTRIBUTIONS::FDE_ELECTROSTATIC_INTERACTIONS,
      ENERGY_CONTRIBUTIONS::FDE_NAD_XC, ENERGY_CONTRIBUTIONS::FDE_NAD_EXACT_EXCHANGE,
      ENERGY_CONTRIBUTIONS::PBE_LINEAR_CORRECTION, ENERGY_CONTRIBUTIONS::FDE_NAD_DISP,
      ENERGY_CONTRIBUTIONS::FDE_FROZEN_SUBSYSTEM_ENERGIES, ENERGY_CONTRIBUTIONS::FDE_ECP_ENV_ENERGY}},
    {ENERGY_CONTRIBUTIONS::MP2_TOTAL_ENERGY, {ENERGY_CONTRIBUTIONS::HF_ENERGY, ENERGY_CONTRIBUTIONS::MP2_CORRECTION}},
    {ENERGY_CONTRIBUTIONS::CCSD_TOTAL_ENERGY, {ENERGY_CONTRIBUTIONS::HF_ENERGY, ENERGY_CONTRIBUTIONS::CCSD_CORRECTION}},
    {ENERGY_CONTRIBUTIONS::CCSD_T_TOTAL_ENERGY,
     {ENERGY_CONTRIBUTIONS::HF_ENERGY, ENERGY_CONTRIBUTIONS::CCSD_CORRECTION, ENERGY_CONTRIBUTIONS::TRIPLES_CORRECTION}}};

const std::map<ENERGY_CONTRIBUTIONS, EnergyContributionParents> EnergyComponentPrinter::ENERGY_CONTRIBUTIONS_PARENTS_MAP =
    determineEnergyContributionParentMap();

const std::map<ENERGY_CONTRIBUTIONS, std::string> EnergyComponentPrinter::_energyContributionMap = {
    {ENERGY_CONTRIBUTIONS::ELECTRON_KINETIC, "Kinetic Energy:             "},
    {ENERGY_CONTRIBUTIONS::ELECTRON_ECP, "ECP Energy:                 "},
    {ENERGY_CONTRIBUTIONS::ELECTRON_NUCLEUS_COULOMB, "Nuclear Attraction:         "},
    {ENERGY_CONTRIBUTIONS::ELECTRON_ELECTRON_COULOMB, "Electron Electron Repulsion (with self-interact.):      "},
    {ENERGY_CONTRIBUTIONS::HF_ELECTRON_ELECTRON_EXCHANGE, "Electron Exchange (HF):          "},
    {ENERGY_CONTRIBUTIONS::KS_DFT_ELECTRON_ELECTRON_EXCHANGE, "Electron Exchange (DFT):          "},
    {ENERGY_CONTRIBUTIONS::ELECTRON_ELECTRON_CORRELATION, "Electron Correlation:       "},
    {ENERGY_CONTRIBUTIONS::NUCLEUS_NUCLEUS_COULOMB, "Nuclear Repulsion:          "},
    {ENERGY_CONTRIBUTIONS::ONE_ELECTRON_ENERGY, "One-Particle Energy (including ECPs):        "},
    {ENERGY_CONTRIBUTIONS::HF_TWO_ELECTRON_ENERGY, "Two-Particle Energy (HF):   "},
    {ENERGY_CONTRIBUTIONS::KS_DFT_TWO_ELECTRON_ENERGY, "Two-Particle Energy (DFT):  "},
    {ENERGY_CONTRIBUTIONS::HF_ENERGY, "Total Energy (HF):          "},
    {ENERGY_CONTRIBUTIONS::KS_DFT_EXCHANGE_CORRELATION_ENERGY, "Exchange-Correlation Energy:"},
    {ENERGY_CONTRIBUTIONS::KS_DFT_EXCHANGE_CORRELATION_NO_HF,
     "Exchange-Correlation Energy (without HF/MP2 contributions):"},
    {ENERGY_CONTRIBUTIONS::KS_DFT_EXACT_EXCHANGE, "Exact exchange (HF exchange in DFT):"},
    {ENERGY_CONTRIBUTIONS::KS_DFT_PERTURBATIVE_CORRELATION, "MP2 correlation (in DFT):"},
    {ENERGY_CONTRIBUTIONS::KS_DFT_ENERGY, "Total Energy (DFT):         "},
    {ENERGY_CONTRIBUTIONS::KS_DFT_DISPERSION_CORRECTION, "Dispersion Correction (DFT):         "},
    {ENERGY_CONTRIBUTIONS::FDE_NAD_EXCHANGE,
     "Non-Additive Exchange Energy (FDE):                                     "},
    {ENERGY_CONTRIBUTIONS::FDE_NAD_CORRELATION,
     "Non-Additive Correlation Energy (FDE):                                  "},
    {ENERGY_CONTRIBUTIONS::FDE_NAD_XC, "Non-Additive Exchange-Correlation Energy (FDE):                         "},
    {ENERGY_CONTRIBUTIONS::FDE_MP2_INT_ENERGY,
     "Non-Additive MP2-Correlation Energy (FDE):                              "},
    {ENERGY_CONTRIBUTIONS::FDE_NAD_EXACT_EXCHANGE,
     "Non-Additive Exact Exchange Energy (FDE):                               "},
    {ENERGY_CONTRIBUTIONS::FDE_NAD_KINETIC, "Non-Additive Kinetic Energy (FDE):                                      "},
    {ENERGY_CONTRIBUTIONS::FDE_NAD_DISP, "Non-Additive Dispersion Correction (FDE):                               "},
    {ENERGY_CONTRIBUTIONS::FDE_ELECTRONS_ENV_ELECTRONS_COULOMB,
     "Electrostatic Interactions of Electrons with Electrons of the Env (FDE):"},
    {ENERGY_CONTRIBUTIONS::FDE_ELECTRONS_ENV_NUCLEI_COULOMB,
     "Electrostatic Interactions of Electrons with Nuclei of the Env (FDE):   "},
    {ENERGY_CONTRIBUTIONS::FDE_NUCLEI_ENV_ELECTRONS_COULOMB,
     "Electrostatic Interactions of Nuclei with Electrons of the Env (FDE):   "},
    {ENERGY_CONTRIBUTIONS::FDE_NUCLEI_ENV_NUCLEI_COULOMB,
     "Electrostatic Interactions of Nuclei with Nuclei of the Env (FDE):      "},
    {ENERGY_CONTRIBUTIONS::FDE_ELECTROSTATIC_INTERACTIONS,
     "Total Electrostatic Interactions (FDE):                                 "},
    {ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_NAD_EXCHANGE,
     "Solv. Non-Additive Exchange Energy (FDE):                               "},
    {ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_NAD_CORRELATION,
     "Solv. Non-Additive Correlation Energy (FDE):                            "},
    {ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_NAD_XC,
     "Solv. Non-Additive Exchange-Correlation Energy (FDE):                   "},
    {ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_NAD_EXACT_EXCHANGE,
     "Solv. Non-Additive Exact Exchange Energy (FDE):                         "},
    {ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_NAD_KINETIC,
     "Solv. Non-Additive Kinetic Energy (FDE):                                "},
    {ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_NAD_DISP,
     "Solv. Non-Additive Dispersion Correction (FDE):                         "},
    {ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_ELECTRONS_ENV_ELECTRONS_COULOMB,
     "Solv. Electrostatic Int. of Electrons with Electrons of the Env (FDE):  "},
    {ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_ELECTRONS_ENV_NUCLEI_COULOMB,
     "Solv. Electrostatic Int. of Electrons with Nuclei of the Env (FDE):     "},
    {ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_NUCLEI_ENV_ELECTRONS_COULOMB,
     "Solv. Electrostatic Int. of Nuclei with Electrons of the Env (FDE):     "},
    {ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_NUCLEI_ENV_NUCLEI_COULOMB,
     "Solv. Electrostatic Int. of Nuclei with Nuclei of the Env (FDE):        "},
    {ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_ELECTROSTATIC_INTERACTIONS,
     "Solv. Total Electrostatic Int. (FDE):                                   "},
    {ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_INTERACTION_ENERGY,
     "Solv. Total Interaction Energy (FDE):                                   "},
    {ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_EMBEDDED_KS_DFT_ENERGY,
     "Solv. Total Energy (embedded DFT):                                      "},
    {ENERGY_CONTRIBUTIONS::FDE_SOLV_SCALED_EMBEDDED_HF_ENERGY,
     "Solv. Total Energy (embedded HF):                                       "},
    {ENERGY_CONTRIBUTIONS::FDE_ECP_ENV_ENERGY,
     "ECP Interaction Energy (FDE):                                           "},
    {ENERGY_CONTRIBUTIONS::FDE_INTERACTION_ENERGY,
     "Total Interaction Energy (FDE):                                         "},
    {ENERGY_CONTRIBUTIONS::FDE_EMBEDDED_KS_DFT_ENERGY,
     "Total Energy (embedded DFT):                                            "},
    {ENERGY_CONTRIBUTIONS::FDE_EMBEDDED_HF_ENERGY,
     "Total Energy (embedded HF):                                             "},
    {ENERGY_CONTRIBUTIONS::FDE_SUPERSYSTEM_ENERGY_DFT_DFT,
     "Total Supersystem Energy (active + all env.):                           "},
    {ENERGY_CONTRIBUTIONS::FDE_SUPERSYSTEM_ENERGY_WF_DFT,
     "Total Supersystem Energy (active + all env.):                           "},
    {ENERGY_CONTRIBUTIONS::FDE_FROZEN_SUBSYSTEM_ENERGIES,
     "Total Frozen Subsystem Energies (environment):                          "},
    {ENERGY_CONTRIBUTIONS::PBE_LINEAR_CORRECTION,
     "Linear Correction to the PBE energy:                                    "},
    {ENERGY_CONTRIBUTIONS::PBE_SUPERSYSTEM_ENERGY_WFT_DFT,
     "Total Supersystem Energy (act. HF + env. DFT/HF):                       "},
    {ENERGY_CONTRIBUTIONS::PBE_SUPERSYSTEM_ENERGY_DFT_DFT,
     "Total Supersystem Energy (act. DFT + env. DFT/HF):                      "},
    {ENERGY_CONTRIBUTIONS::MP2_CORRECTION, "MP2 Energy Correction:                                                  "},
    {ENERGY_CONTRIBUTIONS::CCSD_CORRECTION, "CCSD Energy Correction:                                                 "},
    {ENERGY_CONTRIBUTIONS::TRIPLES_CORRECTION,
     "CCSD(T) Triples Energy Correction:                                      "},
    {ENERGY_CONTRIBUTIONS::MP2_TOTAL_ENERGY,
     "Total MP2 Energy:                                                       "},
    {ENERGY_CONTRIBUTIONS::CCSD_TOTAL_ENERGY,
     "Total CCSD Energy:                                                      "},
    {ENERGY_CONTRIBUTIONS::CCSD_T_TOTAL_ENERGY,
     "Total CCSD(T) Energy:                                                   "},
    {ENERGY_CONTRIBUTIONS::SOLVATION_ENERGY,
     "PCM Solvation Energy:                                                   "},
};
void EnergyComponentPrinter::printEnergyComponent(const std::pair<ENERGY_CONTRIBUTIONS, double>& dataToPrint) {
  if (_energyContributionMap.find(dataToPrint.first) == _energyContributionMap.end())
    throw SerenityError(
        "ERROR: An energy contribution is supposed to be printed to the output which does not have any output-name!");
  printf("%-80s %18.10f \n", _energyContributionMap.at(dataToPrint.first).c_str(), dataToPrint.second);
}

const std::map<ENERGY_CONTRIBUTIONS, EnergyContributionParents> EnergyComponentPrinter::determineEnergyContributionParentMap() {
  std::map<ENERGY_CONTRIBUTIONS, std::vector<ENERGY_CONTRIBUTIONS>> parentMap;
  for (const auto& energyType : ENERGY_CONTRIBUTIONS_VECTOR) {
    std::vector<ENERGY_CONTRIBUTIONS> parents;
    for (const auto& possibleParentType : ENERGY_CONTRIBUTIONS_CHILDREN_MAP) {
      for (const auto& childOfPossibleParent : possibleParentType.second) {
        if (childOfPossibleParent == energyType) {
          // One son/daughter of this dad/mom is me, => this dad/mom is actually my dad/mom.
          parents.push_back(possibleParentType.first);
          break;
        }
      }
    }
    parentMap.emplace(energyType, parents);
  }
  return parentMap;
}

bool EnergyComponentPrinter::isAncestorOf(const ENERGY_CONTRIBUTIONS& descendant, const ENERGY_CONTRIBUTIONS& ancestor) {
  if (descendant == ancestor)
    return true;
  for (const auto& child : ENERGY_CONTRIBUTIONS_CHILDREN_MAP.at(ancestor)) {
    if (child == descendant || isAncestorOf(descendant, child))
      return true;
  }
  return false;
}

} /* namespace Serenity */
